#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['trvaep', 'trvaep.model']

package_data = \
{'': ['*'], 'trvaep.model': ['.ipynb_checkpoints/*']}

install_requires = \
['future-fstrings',
 'get_version',
 'get_version',
 'scanpy',
 'adjustText',
 'torch',
 'torchvision',
 'numpy',
 'scipy',
 'seaborn',
 'matplotlib',
 'pandas']

extras_require = \
{'test': ['pytest']}

setup(name='trvaep',
      version='0.0.9',
      description='trVAE - Transfer Variational Autoencoders pytorch',
      author='Mohammad Lotfollahi',
      author_email='mohammad.lotfollahi@helmholtz-muenchen.de',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
